#!/usr/bin/env python

import time
import sys
from array import array

def fileLen(fname):
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1

def init(buffer):
   from glob import globs, setGlobals
   from array import array

   # create spatial variables of specified number
   temp = buffer.pop(0)
   Vars = int(temp.split()[0].strip())+2
   TypeId = temp.split()[1].strip()
   
   # initialize constraints
   N = fileLen(globs['calc']+"/"+globs['calc']+".relations")
   setGlobals("size", N)
   Id = 0
   from bitcoding import B_dict
   with open(globs['calc']+"/"+globs['calc']+".identity") as f:
      Id = B_dict[f.readline().strip()]
   if N <= 8:
      ConMatrix = tuple([array('B',[B_dict['DALL'] if i != j else Id for i in xrange(Vars)]) for j in xrange(Vars)])
   elif N <= 16:
      ConMatrix = tuple([array('H',[B_dict['DALL'] if i != j else Id for i in xrange(Vars)]) for j in xrange(Vars)])
   else:
      ConMatrix = tuple([array('I',[B_dict['DALL'] if i != j else Id for i in xrange(Vars)]) for j in xrange(Vars)])

   slimConMatrix = [None for v in xrange(Vars)]
   for v in xrange(Vars):
      j = {}   
      slimConMatrix[v] = j

   # parse spatial CSP
   from parsecsp import parsecsp
   parsecsp(ConMatrix, buffer, slimConMatrix)
   
   return TypeId, ConMatrix, slimConMatrix

class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg

# main function
def main(argv=None):

   if argv is None:
        argv = sys.argv

   # set some sys useful stuff
   sys.setrecursionlimit(1<<30)
   sys.setcheckinterval(10000)

   from optparse import OptionParser

   # initialize parser for command line arguments
   parser = OptionParser()
   
   # set parsing options
   parser.add_option("-c", "--calculus", type="string", dest="calc", default="rcc8", help="interaction mode: rcc8, or allen [default: %default]")
   parser.add_option("-p", "--pcmethod", type="string", dest="pcmethod", default="full", help="interaction mode: full, or incremental [default: %default]")
   parser.add_option("-d", "--print", action="store_true", dest="printsol", help="print solution to output")
   parser.add_option("-f", "--file", action="store", type="string", dest="filename")

   # parse command line arguments
   options, args = parser.parse_args()

   from glob import setGlobals
   #import argparse

   try:   
      # check and set globals for calculus information
      if not (options.calc == "rcc8" or options.calc == "allen"):
         raise Usage("Invalid option: " + options.calc)
      else:
         setGlobals('calc',options.calc)

      # check and set globals for patch consistency heuristic information
      if options.pcmethod == "full":
         setGlobals('pcmethod',0)
      elif options.pcmethod == "incremental":
         setGlobals('pcmethod',1)
      else:
         raise Usage("Invalid option: " + options.pcheuristic)

   except Usage, err:
        print err.msg
        parser.print_help()
        return 1

   with open(options.filename) as f:
      total = 0
      totalc = 0
      totali = 0
      meter = 0
      meteri = 0
      edges = 0
      edgesc = 0
      edgesi = 0
      cons = 0
      consc = 0
      consi = 0
      buffer = []

      for i in f:
         buffer.append(i)

         if i.strip() == '.':
            start = elapsed = 0
            meter += 1

            from counters import counter

            TypeId, ConMatrix, slimConMatrix = init(buffer)
            buffer = []

            neighbors = {}
                  
            from glob import globs
            if globs['pcmethod'] == 0: # path consistency
               from pc import PC as pc

               start = time.clock()
               solution = pc(ConMatrix,neighbors, slimConMatrix)
               elapsed = time.clock() - start
               if solution:
                  totalc += elapsed
               else:
                  totali += elapsed
                  meteri += 1 
               total += elapsed

            if globs['pcmethod'] == 1: # incremental path consistency
               from ipc import iPC as pc

               elapsed = 0
                          
               for step in xrange(len(ConMatrix)): 

                  elapsed2 = 0
 
                  start = time.clock()
                  solution = pc(ConMatrix,neighbors,step, slimConMatrix)
                  elapsed2 += time.clock() - start
                  elapsed += elapsed2
                  total += elapsed2

                  if not solution: 
                     totali += elapsed
                     meteri += 1
                     break
               else:
                  totalc += elapsed

            # print values of complexity counters
            print '--------------------------------------'
            print 'Type ID: ', TypeId
            print '--------------------------------------'
            print 'Visited arcs: ', counter['arcs']
            print 'Checked constraints: ', counter['cons']
            print '--------------------------------------'
            print 'Time: ', elapsed, 'sec'
            print '--------------------------------------'

            # print solution  
            if solution != False:
               edgesc += counter['arcs']
               consc += counter['cons']
               print 'Path Consistent'
               if options.printsol == True:
                  from printm import printMatrix
                  printMatrix(ConMatrix)
            else:
               edgesi += counter['arcs']
               consi += counter['cons']
               print 'Inconsistent'
            print '--------------------------------------'

            edges += counter['arcs']
            cons += counter['cons']
            counter['arcs'] = 0
            counter['cons'] = 0
            ConMatrix = None
          
   print '======================================'
   print 'Total time:', total, 'sec'
   print 'Avg time per CSP:', total/(1.0*meter), 'sec'
   if meter-meteri:
      print 'Avg time per path consistent CSP:', totalc/(1.0*(meter-meteri)), 'sec' 
   if meteri:
      print 'Avg time per inconsistent CSP:', totali/(1.0*meteri), 'sec'
   print 'Avg # of arcs processed per CSP:', edges/(1.0*meter)
   if meter-meteri:
      print 'Avg # of arcs processed per path consistent CSP:', edgesc/(1.0*(meter-meteri))
   if meteri:
      print 'Avg # of arcs processed per inconsistent CSP:', edgesi/(1.0*meteri)
   print 'Avg # of constraint checks per CSP:', cons/(1.0*meter)
   if meter-meteri:
      print 'Avg # of constraint checks per path consistent CSP:', consc/(1.0*(meter-meteri)) 
   if meteri:
      print 'Avg # of constraint checks per inconsistent CSP:', consi/(1.0*meteri)
   if meter-meteri:
      print '# of path consistent CSPs:', meter-meteri
   print '# of inconsistent CSPs:', meteri
   print '======================================'

if __name__ == '__main__':
   sys.exit(main())

